import colorama

# color is reset at the end of every print
colorama.init(autoreset=True)

# ------ colors definitions

info = colorama.Fore.LIGHTWHITE_EX
reset = colorama.Style.RESET_ALL
error = colorama.Fore.RED
extra_info = colorama.Fore.LIGHTCYAN_EX
